#pragma once

#include "..\..\common\util_defs.h"

/* NeuQuant Neural-Net Quantization Algorithm Interface
 * ----------------------------------------------------
 *
 * Copyright (c) 1994 Anthony Dekker
 *
 * NEUQUANT Neural-Net quantization algorithm by Anthony Dekker, 1994.
 * See "Kohonen neural networks for optimal colour quantization"
 * in "Network: Computation in Neural Systems" Vol. 5 (1994) pp 351-367.
 * for a discussion of the algorithm.
 * See also  http://members.ozemail.com.au/~dekker/NEUQUANT.HTML
 *
 * Any party obtaining a copy of these files from the author, directly or
 * indirectly, is granted, free of charge, a full and unrestricted irrevocable,
 * world-wide, paid up, royalty-free, nonexclusive right and license to deal
 * in this software and documentation files (the "Software"), including without
 * limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons who receive
 * copies from any such party to do so, with the only requirement being
 * that this copyright notice remain intact.
 */


#include <stdio.h>

/* four primes near 500 - assume no image has a length so large */
/* that it is divisible by all four primes */
#define prime1		499
#define prime2		491
#define prime3		487
#define prime4		503
#define minpicturebytes	(3*prime4)		/* minimum size for input image */
#define netbiasshift	4			/* bias for colour values */
#define ncycles		100			/* no. of learning cycles */

/* defs for freq and bias */
#define intbiasshift    16			/* bias for fractions */
#define intbias		(((int) 1)<<intbiasshift)
#define gammashift  	10			/* gamma = 1024 */
#define gamma   	(((int) 1)<<gammashift)
#define betashift  	10
#define beta		(intbias>>betashift)	/* beta = 1/1024 */
#define betagamma	(intbias<<(gammashift-betashift))

/* defs for decreasing radius factor */
#define radiusbiasshift	6			/* at 32.0 biased by 6 bits */
#define radiusbias	(((int) 1)<<radiusbiasshift)
#define radiusdec	30			/* factor of 1/30 each cycle */ 

/* defs for decreasing alpha factor */
#define alphabiasshift	10			/* alpha starts at 1.0 */
#define initalpha	(((int) 1)<<alphabiasshift)

/* radbias and alpharadbias used for radpower calculation */
#define radbiasshift	8
#define radbias		(((int) 1)<<radbiasshift)
#define alpharadbshift  (alphabiasshift+radbiasshift)
#define alpharadbias    (((int) 1)<<alpharadbshift)

typedef int pixel[4];				/* BGRc */


class NeuQuant{
public:
	NeuQuant(int net_size, int sample_size, int stride_, int width_, int height_)
	{ 
		SetNewParameters(net_size, sample_size); 
		stride = stride_; 
		height = height_;
		width  = width_;
	}

	void SetNewParameters(int net_size, int sample_size)
	{ 
		netsize    = net_size;
		maxnetpos  = net_size - 1;
		initrad    = net_size >> 3;
		initradius = initrad * radiusbias;
		samplefac  = sample_size;
		run_previously = false;
	}

	// Initialise network in range (0,0,0) to (255,255,255) and set parameters
	void initnet(unsigned char *thepic, int len);
			
	// Unbias network to give byte values 0..255 and record position i to prepare for sort
	void unbiasnet();	/* can edit this function to do output of colour map */

	void putcolourmaptomem(byte* mem);

	// Insertion sort of network and building of netindex[0..255] (to do after unbias)
	void inxbuild();

	// Search for BGR values 0..255 (after net is unbiased) and return colour index
	int inxsearch(int b, int g, int r);

	// Main Learning Loop
	void learn();

	// Marshall neuron-network
	bool Marshall(byte*& output, int& length);

	// DeMarshall neuron-network
	bool DeMarshall(byte* output, int length);

private:
	// Search for biased BGR values
	int contest(int b, int g, int r);

	// Move neuron i towards biased (b,g,r) by factor alpha
	void altersingle(int alpha, int i, int b, int g, int r);

	// Move adjacent neurons by precomputed alpha*(1-((i-j)^2/[r]^2)) in radpower[|i-j|]
	void alterneigh(int rad, int i, int b, int g, int r);

public:
	bool	run_previously;

private:
	int netsize;			/* number of colours used */
	int maxnetpos;
	int initrad;			/* for 256 cols, radius starts */
	int initradius;			/* and decreases by a */
	int alphadec;			/* biased by 10 bits */
	int radpower[64];		/* radpower for precomputation */

	unsigned char *thepicture;		/* the input image itself */
	int		lengthcount;			/* lengthcount = H*W*3 */
	int		samplefac;				/* sampling factor 1..30 */
	int		netindex[256];			/* for network lookup - really 256 */
	int		bias[256];				/* bias and freq arrays for learning */
	int		freq[256];
	pixel	network[256];			/* the network itself */
	int		stride, width, height;
};



/* Program Skeleton
   ----------------
   	[select samplefac in range 1..30]
   	pic = (unsigned char*) malloc(3*width*height);
   	[read image from input file into pic]
	initnet(pic,3*width*height,samplefac);
	learn();
	unbiasnet();
	[write output image header, using writecolourmap(f),
	possibly editing the loops in that function]
	inxbuild();
	[write output image using inxsearch(b,g,r)]		*/
